from agentine.llm import LLM, Message, FileMessage
from agentine.agent import Agent, AgentGroup, AgentIndex, AgentLegion
from agentine.chatbot import Chatbot
from .config import (
    DEFAULT_LLM_PROVIDER,
    KNOWN_LLM_PROVIDERS,
)


import os
from dotenv import load_dotenv
load_dotenv()

DEFAULT_LLM_PROVIDER = "openai"

OPENAI = {
    "api_key": os.getenv("OPENAI_API_KEY", ""),
    "base_url": "https://api.openai.com/v1",
    "default_model": "gpt-5-chat-latest",
}

GROQ = {
    "api_key": os.getenv("GROQ_API_KEY", ""),
    "base_url": "https://api.groq.com/openai/v1",
    "default_model": "llama-3.3-70b-versatile",
}

GOOGLE = {
    "api_key": os.getenv("GEMINI_API_KEY", ""),
    "base_url": "https://generativelanguage.googleapis.com/v1beta/openai/",
    "default_model": "gemini-2.5-flash-lite",
}

PERPLEXITY = {
    "api_key": os.getenv("PERPLEXITY_API_KEY", ""),
    "base_url": "https://api.perplexity.ai",
    "default_model": "sonar",
}

ANTHROPIC = {
    "api_key": os.getenv("ANTHROPIC_API_KEY", ""),
    "base_url": "https://api.anthropic.com",
    "default_model": "claude-4-sonnet-latest",
}

XAI = {
    "api_key": os.getenv("XAI_API_KEY", ""),
    "base_url": "https://api.x.ai/v1",
    "default_model": "grok-3-mini",
}

DEEPSEEK = {
    "api_key": os.getenv("DEEPSEEK_API_KEY", ""),
    "base_url": "https://api.deepseek.com",
    "default_model": "deepseek-chat",
}

MISTRAL = {
    "api_key": os.getenv("MISTRAL_API_KEY", ""),
    "base_url": "https://api.mistral.ai/v1",
    "default_model": "mistral-medium",
}

COHERE = {
    "api_key": os.getenv("COHERE_API_KEY", ""),
    "base_url": "https://api.cohere.ai/compatibility/v1",
    "default_model": "command-r",
}

KNOWN_LLM_PROVIDERS = {
    "openai": OPENAI,
    "groq": GROQ,
    "google": GOOGLE,
    "perplexity": PERPLEXITY,
    "anthropic": ANTHROPIC,
    "xai": XAI,
    "deepseek": DEEPSEEK,
    "mistral": MISTRAL,
    "cohere": COHERE,
}


import time
from agentine.llm import LLM, Message

def test_llm_init(provider: str = "openai") -> float:
    start = time.perf_counter()
    llm = LLM(provider=provider)
    end = time.perf_counter()
    elapsed = end - start
    print(f"LLM init for provider '{provider}' took {elapsed:.4f} seconds")
    return elapsed

def test_llm_switch(initial_provider: str = "openai", new_provider: str = "groq") -> float:
    llm = LLM(provider=initial_provider)
    print(f"Initialized with provider '{initial_provider}' and model '{llm.model}'")

    start = time.perf_counter()
    llm.provider = new_provider  # triggers __setattr__ and re-init of clients
    end = time.perf_counter()

    elapsed = end - start
    print(f"Switched to provider '{new_provider}' with model '{llm.model} {llm.base_url} {llm.api_key}' in {elapsed:.4f} seconds")
    return elapsed

# # Test OpenAI
# test_llm_init("openai")

# # Test Groq
# test_llm_init("groq")

# # Test Anthropic
# test_llm_init("anthropic")

# Switch from OpenAI to Groq
# test_llm_switch("openai", "gemini")

# # Switch from Groq to Anthropic
# test_llm_switch("groq", "openai")

# # Switch from OpenAI to DeepSeek
# test_llm_switch("openai", "deepseek")

llm = LLM()
llm.provider = "google"
# print("New config:")
# print(llm)
print(llm.chat([Message(content="Who is the first person walking on the moon?")]))


from typing import Dict, Any, Optional, List, Union, Iterator, AsyncIterator
from pydantic import BaseModel, Field, PrivateAttr
from openai import OpenAI, AsyncOpenAI
import asyncio
from concurrent.futures import ThreadPoolExecutor

from agentine.config import DEFAULT_LLM_PROVIDER, KNOWN_LLM_PROVIDERS
from agentine.message import Message, FileMessage


class LLM(BaseModel):
    api_key: Optional[str] = Field(default=None)
    provider: str = Field(default=DEFAULT_LLM_PROVIDER)
    model: Optional[str] = None
    base_url: str = ""

    timeout: int = 60000
    max_retries: int = 2
    max_concurrency: int = 100

    response_format: str = Field(default="text")
    temperature: float = 1.0
    max_completion_tokens: Optional[int] = None
    reasoning_effort: Optional[str] = None

    _client: OpenAI = PrivateAttr()
    _client_async: AsyncOpenAI = PrivateAttr()

    class Config:
        extra = "allow"

    def __init__(self, **data: Any):
        if "provider" not in data:
            data["provider"] = DEFAULT_LLM_PROVIDER

        super().__init__(**data)
        self._apply_provider_config(self.provider)
        self._init_clients()

    def __setattr__(self, name: str, value: Any) -> None:
        super().__setattr__(name, value)

        if name == "provider":
            self._apply_provider_config(value)

        if name in {"provider", "api_key", "base_url"}:
            self._update_clients()

    def _apply_provider_config(self, provider: str) -> None:
        provider_key = str(provider).lower()
        provider_config = KNOWN_LLM_PROVIDERS.get(provider_key)

        if not provider_config:
            raise ValueError(
                f"422 - UNKNOWN PROVIDER: '{provider.upper()}'\n\n"
                f"Valid providers are: {', '.join(KNOWN_LLM_PROVIDERS.keys())}\n"
                f"If you want to use a custom provider, please set 'api_key' and 'base_url' manually instead."
            )

        super().__setattr__("model", provider_config.get("default_model"))
        super().__setattr__("api_key", provider_config.get("api_key"))
        super().__setattr__("base_url", provider_config.get("base_url"))

    def _init_clients(self) -> None:
        self._client = OpenAI(api_key=self.api_key, base_url=self.base_url)
        self._client_async = AsyncOpenAI(api_key=self.api_key, base_url=self.base_url)

    def _update_clients(self) -> None:
        self._client.api_key = self.api_key
        self._client.base_url = self.base_url
        self._client_async.api_key = self.api_key
        self._client_async.base_url = self.base_url

    @property
    def client(self) -> OpenAI:
        return self._client

    @property
    def client_async(self) -> AsyncOpenAI:
        return self._client_async

    def completion_config(self) -> Dict[str, Any]:
        excluded = {
            "api_key",
            "provider",
            "base_url",
            "timeout",
            "max_retries",
            "max_concurrency",
        }
        data = self.model_dump(exclude=excluded, exclude_none=True)
        if "response_format" in data:
            data["response_format"] = {"type": data["response_format"]}
        return data

    def _prepare_kwargs(self, messages: List["Message"]) -> Dict[str, Any]:
        assert isinstance(messages, list) and all(isinstance(m, Message) for m in messages)
        kwargs = self.completion_config()
        kwargs["model"] = self.model
        kwargs["messages"] = [m.core() for m in messages]
        return kwargs

    def chat(self, messages: List["Message"]) -> "Message":
        kwargs = self._prepare_kwargs(messages)
        for attempt in range(self.max_retries + 1):
            try:
                response = self.client.chat.completions.create(**kwargs)
                return Message.from_openai_completion(response)
            except Exception:
                if attempt == self.max_retries:
                    raise

    async def chat_async(self, messages: List["Message"]) -> "Message":
        kwargs = self._prepare_kwargs(messages)
        for attempt in range(self.max_retries + 1):
            try:
                response = await self.client_async.chat.completions.create(**kwargs)
                return Message.from_openai_completion(response)
            except Exception:
                if attempt == self.max_retries:
                    raise

    def stream(
        self, messages: List["Message"], include_usage: bool = True
    ) -> Iterator["Message"]:
        kwargs = self._prepare_kwargs(messages)
        kwargs["stream"] = True
        if include_usage:
            kwargs["stream_options"] = {"include_usage": True}
        stream = self.client.chat.completions.create(**kwargs)
        for chunk in stream:
            yield Message.from_openai_completion_chunk(chunk)

    async def stream_async(
        self, messages: List["Message"], include_usage: bool = True
    ) -> AsyncIterator["Message"]:
        kwargs = self._prepare_kwargs(messages)
        kwargs["stream"] = True
        if include_usage:
            kwargs["stream_options"] = {"include_usage": True}
        stream = await self.client_async.chat.completions.create(**kwargs)
        async for chunk in stream:
            yield Message.from_openai_completion_chunk(chunk)

    def batch(self, batch_messages: List[List["Message"]]) -> List[Union["Message", Exception]]:
        def process(messages: List["Message"]) -> Union["Message", Exception]:
            try:
                return self.chat(messages)
            except Exception as e:
                return e
        with ThreadPoolExecutor(max_workers=self.max_concurrency) as executor:
            return list(executor.map(process, batch_messages))

    async def batch_async(self, batch_messages: List[List["Message"]]) -> List[Union["Message", Exception]]:
        async def process(messages: List["Message"]) -> Union["Message", Exception]:
            try:
                return await self.chat_async(messages)
            except Exception as e:
                return e
        return await asyncio.gather(*(process(m) for m in batch_messages), return_exceptions=True)


from agentine.agent import Agent
from agentine.chatbot import Chatbot
from agentine.llm import LLM
from agentine.llm import Message


general_agent = Agent(
    instruction="You are a helpful assistant. Answer the user's questions clearly and concisely.",
)
general_agent.llm.provider = "openai"
# general_agent.llm.provider = "groq"
# general_agent.llm.provider = "google"
# general_agent.llm.provider = "xai"
# general_agent.llm.provider = "openai"


# result = general_agent.work(query="hello, how are you today?", messages=[])
# print(result)

chatbot = Chatbot(client=general_agent)
chatbot.cli_run(stream=False, display_stats=True)


import json
from typing import AsyncIterator, Iterator, List, Optional, Dict, Any, Tuple, Literal
from pydantic import BaseModel, Field
import asyncio
from agentine.llm import LLM, Message
from agentine.utils import Utility

class Agent(BaseModel):
    instruction: str
    name: Optional[str] = None
    llm: LLM = Field(default_factory=LLM)
    response_format: Literal["text", "json_schema", "json_object"] = Field(default="text")

    class Config:
        extra = "allow"

    def __init__(self, **data: Any) -> None:
        super().__init__(**data)
        self.llm.response_format = self.response_format

    def __setattr__(self, key: str, value: Any) -> None:
        super().__setattr__(key, value)
        if key == "response_format" and hasattr(self, "llm") and self.llm is not None:
            self.llm.response_format = value

    def _prepare_messages(
        self,
        query: Optional[str] = None,
        messages: Optional[List[Message]] = None,
        data: Optional[Dict[str, Any]] = None,
    ) -> List[Message]:
        formatted_instruction = Utility.format(string=self.instruction, data=data) if data else self.instruction
        formatted_query = Utility.format(string=query, data=data) if data else query

        chat_messages: List[Message] = []

        if formatted_instruction:
            chat_messages.append(Message(role="system", content=formatted_instruction))
        if messages:
            chat_messages.extend(messages)
        if formatted_query:
            chat_messages.append(Message(role="user", content=formatted_query))

        return chat_messages

    def work(
        self,
        query: Optional[str] = None,
        messages: Optional[List[Message]] = None,
        data: Optional[Dict[str, Any]] = None,
    ) -> Message:
        assert query is None or isinstance(query, str), "query must be a string or None"
        assert messages is None or isinstance(messages, list), "messages must be a list or None"
        assert data is None or isinstance(data, dict), "data must be a dict or None"

        chat_messages = self._prepare_messages(query=query, messages=messages, data=data)
        result = self.llm.chat(messages=chat_messages)

        if self.response_format == "json_object":
            result.data = json.loads(result.content)

        return result

    async def work_async(
        self,
        query: Optional[str] = None,
        messages: Optional[List[Message]] = None,
        data: Optional[Dict[str, Any]] = None,
    ) -> Message:
        assert query is None or isinstance(query, str), "query must be a string or None"
        assert messages is None or isinstance(messages, list), "messages must be a list or None"
        assert data is None or isinstance(data, dict), "data must be a dict or None"

        chat_messages = self._prepare_messages(query=query, messages=messages, data=data)
        result = await self.llm.chat_async(messages=chat_messages)

        if self.response_format == "json_object":
            result.data = json.loads(result.content)

        return result

    def stream(
        self,
        query: Optional[str] = None,
        messages: Optional[List[Message]] = None,
        data: Optional[Dict[str, Any]] = None,
    ) -> Iterator[Message]:
        assert query is None or isinstance(query, str), "query must be a string or None"
        assert messages is None or isinstance(messages, list), "messages must be a list or None"
        assert data is None or isinstance(data, dict), "data must be a dict or None"

        chat_messages = self._prepare_messages(query=query, messages=messages, data=data)
        return self.llm.stream(chat_messages)

    async def stream_async(
        self,
        query: Optional[str] = None,
        messages: Optional[List[Message]] = None,
        data: Optional[Dict[str, Any]] = None,
    ) -> AsyncIterator[Message]:
        assert query is None or isinstance(query, str), "query must be a string or None"
        assert messages is None or isinstance(messages, list), "messages must be a list or None"
        assert data is None or isinstance(data, dict), "data must be a dict or None"

        chat_messages = self._prepare_messages(query=query, messages=messages, data=data)
        async for message in self.llm.stream_async(chat_messages):
            yield message

class AgentGroup(BaseModel):
    agents: List[Agent] = Field(default_factory=list)

    def work(
        self,
        query: str,
        messages: Optional[List[Message]] = None,
        data: Optional[Dict[str, Any]] = None,
    ) -> List[Message]:
        results = []
        for agent in self.agents:
            result = agent.work(query=query, messages=messages, data=data)
            results.append(result)
        return results

    async def work_async(
        self,
        query: str,
        messages: Optional[List[Message]] = None,
        data: Optional[Dict[str, Any]] = None,
    ) -> List[Message]:
        tasks = [
            agent.work_async(query=query, messages=messages, data=data)
            for agent in self.agents
        ]
        results = await asyncio.gather(*tasks)
        return results

class AgentIndex(BaseModel):
    agents: Dict[str, Agent] = Field(default_factory=dict)
    default: Optional[Agent] = None

    def __getitem__(self, key: str) -> Agent:
        return self.agents[key]

    def add(self, name: str, agent: Agent, is_default: bool = False) -> None:
        self.agents[name] = agent
        if is_default:
            self.default = agent

    def remove(self, name: str) -> None:
        if name not in self.agents:
            raise KeyError(f"Agent '{name}' not found in the index.")

        if self.default == self.agents[name]:
            self.default = None

        del self.agents[name]

    def set_default(self, name: str) -> None:
        if name not in self.agents:
            raise KeyError(f"Agent '{name}' not found in the index.")
        self.default = self.agents[name]

    def find(self, names: List[str]) -> List[Agent]:
        return [agent for name in names if (agent := self.agents.get(name)) is not None]

class AgentLegion(BaseModel):
    speaker: Agent
    selector: Agent
    agent_index: AgentIndex

    def _prepare_messages(
        self,
        query: Optional[str] = None,
        messages: Optional[List[Message]] = None,
    ) -> Tuple[List[Message], Optional[Agent]]:
        if messages is None:
            messages = []

        selector_result = self.selector.work(query=query, messages=messages)

        assert selector_result.data is not None, "Selector result missing data."
        assert "selections" in selector_result.data, "Selector data missing 'selections' key."

        selections = selector_result.data["selections"]
        selected_agents = self.agent_index.find(selections)

        if len(selected_agents) == 1:
            return messages, selected_agents[0]

        agent_group = AgentGroup(agents=selected_agents)

        results = asyncio.run(
            agent_group.work_async(query=query, messages=messages)
        )

        for agent_name, result in zip(selections, results):
            result.content = (
                f"**{agent_name} agent** response (NOT VISIBLE TO USER):\n\n{result.content}"
            )
            messages.append(result)

        return messages, None

    def work(
        self,
        query: Optional[str] = None,
        messages: Optional[List[Message]] = None,
    ) -> Message:
        messages, single_agent = self._prepare_messages(query=query, messages=messages)

        if single_agent is not None:
            return single_agent.work(query=query, messages=messages)

        return self.speaker.work(messages=messages)

    def stream(
        self,
        query: Optional[str] = None,
        messages: Optional[List[Message]] = None,
    ) -> Iterator[Message]:
        messages, single_agent = self._prepare_messages(query=query, messages=messages)

        if single_agent is not None:
            yield from single_agent.stream(query=query, messages=messages)
            return

        yield from self.speaker.stream(messages=messages)


from typing import List, Any
from pydantic import BaseModel, Field
from agentine.llm import Message, FileMessage
from agentine.utils import Utility
from agentine.agent import Agent
from print9 import print9


class Chatbot(BaseModel):
    client: Any
    messages: List[Message] = Field(default_factory=list)

    def cli_run(self, stream: bool = False, display_stats: bool = False):
        print("Chatbot started. Type 'exit' to quit.")
        print()
        while True:
            query = input("YOU: ")
            if query.lower() == "exit":
                print("Chatbot session ended.")
                break

            if query == "--upload file":
                file_message = FileMessage.from_terminal()
                print9(f"{len(file_message.files)} images uploaded.")
                text = input("YOU: ")
                file_message.text = text
                self.messages.append(file_message)
            else:
                user_message = Message(content=query)
                self.messages.append(user_message)

            if stream:
                print()
                print9("BOT: ", end="", color="green", flush=True)
                accumulated_content = ""
                stats = {}
                for chunk in self.client.stream(messages=self.messages):
                    if chunk.content is not None:
                        print9(chunk.content, color="green", end="", flush=True)
                        accumulated_content += chunk.content
                    stats = chunk.stats

                print("\n")
                full_response = Message(
                    role="assistant",
                    content=accumulated_content,
                    stats=stats,
                )
                self.messages.append(full_response)

                if display_stats:
                    print9(stats, color="red")
                    print()
            else:
                response = self.client.work(messages=self.messages)
                self.messages.append(response)
                print()
                print9(f"BOT: {response.content}", color="green")
                print()

                if display_stats:
                    print9(response.stats, color="red")
                    print()


from datetime import datetime
from typing import List, Any
from pydantic import BaseModel, Field

class ChangeLog(BaseModel):
    time: datetime = Field(default_factory=datetime.utcnow)
    fields: List[str] = Field(default_factory=list, min_items=1)

class Metadata(BaseModel):
    created_at: datetime = Field(default_factory=datetime.utcnow)
    change_logs: List[ChangeLog] = Field(default_factory=list)

    class Config:
        extra = "allow"

    def __setattr__(self, name: str, value: Any) -> None:
        if name == "created_at":
            if hasattr(self, name):
                raise ValueError(f"'{name}' field is immutable and cannot be changed.")
        super().__setattr__(name, value)

    def log_change(self, fields: List[str]) -> None:
        self.change_logs.append(ChangeLog(fields=fields))

class Stats(BaseModel):
    class Config:
        extra = "allow"


from typing import Dict, Any, Optional, List, Union, Literal
from pydantic import BaseModel, Field
import base64
import mimetypes

from agentine.metadata import Metadata, Stats
from agentine.utils import Utility


class Message(BaseModel):
    role: Optional[Literal["system", "developer", "user", "assistant", "tool"]] = "user"
    content: Optional[Union[str, List[Dict[str, Any]]]] = None
    data: Optional[Any] = None

    stats: Stats = Field(default_factory=Stats)
    metadata: Metadata = Field(default_factory=Metadata)

    class Config:
        extra = "allow"

    def __setattr__(self, name: str, value: Any) -> None:
        if name != "metadata":
            old_value = getattr(self, name, None)
            super().__setattr__(name, value)
            if old_value != value:
                self.metadata.log_change(fields=[name])
        else:
            super().__setattr__(name, value)

    def core(self) -> Dict[str, str]:
        return {"role": self.role, "content": self.content}

    @classmethod
    def from_openai_completion(cls, completion: Any) -> "Message":
        completion_dict = (
            completion.model_dump()
            if hasattr(completion, "model_dump")
            else dict(completion)
        )
        usage_dict = completion_dict.pop("usage", None)
        choices = completion_dict.pop("choices", [])
        choice_dict = choices[0] if choices else {}
        message = choice_dict.get("message", {})
        content = message.pop("content", "") if isinstance(message, dict) else ""
        message.pop("role", None)
        role = "assistant"

        return cls(
            content=content,
            role=role,
            stats=Stats(choice=choice_dict, completion=completion_dict, usage=usage_dict),
        )
        
    @classmethod
    def from_openai_completion_chunk(cls, chunk: Any) -> "Message":
        chunk_dict = (
            chunk.model_dump() if hasattr(chunk, "model_dump") else dict(chunk)
        )

        usage_dict = chunk_dict.pop("usage", None)
        choices = chunk_dict.pop("choices", [])
        choice_dict = choices[0] if choices else {}
        delta = choice_dict.get("delta", {})
        content = delta.pop("content", "") if isinstance(delta, dict) else ""
        delta.pop("role", None)
        role = "assistant"
        message = cls(
            role=role,
            content=content,
            stats=Stats(choice=choice_dict, completion=chunk_dict, usage=usage_dict),
        )
        message.metadata.is_chunk = True

        return message


class FileMessage(Message):
    text: str = ""
    files: List[Dict[str, Any]] = Field(default_factory=list)
    content: List[Dict] = Field(default_factory=list)

    def __init__(self, **data):
        super().__init__(**data)
        self._update_content()

    def __setattr__(self, name, value):
        super().__setattr__(name, value)
        if name in {"text", "files"}:
            self._update_content()

    def _update_content(self):
        content_blocks = []
        if self.text:
            content_blocks.append({"type": "text", "text": self.text})

        for file_info in self.files:
            filename = file_info.get("filename")
            data_url = file_info.get("data_url")
            mime_type = file_info.get("mime_type", "application/octet-stream")

            if mime_type.startswith("image/"):
                content_blocks.append(
                    {"type": "image_url", "image_url": {"url": data_url}}
                )
            else:
                content_blocks.append(
                    {
                        "type": "file",
                        "file": {
                            "filename": filename,
                            "file_data": data_url,
                        },
                    }
                )

        super().__setattr__("content", content_blocks)

    def core(self) -> dict:
        return {"role": self.role, "content": self.content}

    @staticmethod
    def from_terminal(text: str = "") -> "FileMessage":
        file_path = Utility.get_file_path_via_terminal()
        if not file_path:
            raise ValueError("No file selected")

        with open(file_path, "rb") as f:
            file_bytes = f.read()

        filename = file_path.split("/")[-1]
        mime_type, _ = mimetypes.guess_type(filename)
        if mime_type is None:
            mime_type = "application/octet-stream"

        b64_str = base64.b64encode(file_bytes).decode("utf-8")
        data_url = f"data:{mime_type};base64,{b64_str}"

        files_list = [
            {
                "filename": filename,
                "data_url": data_url,
                "mime_type": mime_type,
            }
        ]

        return FileMessage(text=text, files=files_list)


import re
import textwrap
from typing import Dict, Any, Literal, Optional, List, Union
import tkinter as tk
from tkinter import filedialog
from print9 import print9

class Utility:
    
    @staticmethod
    def format(
        string: str,
        data: Dict[str, Any],
        fallback: str = "(Not Available)"
    ) -> str:
        assert isinstance(string, str), f"string must be a string. Value: {string!r}"
        assert isinstance(data, dict), f"data must be a dictionary. Value: {data!r}"
        assert isinstance(fallback, str), f"fallback must be a string. Value: {fallback!r}"

        placeholders = re.findall(r'<<(.*?)>>', string)

        for key in set(placeholders):
            value = data.get(key, fallback)
            if not isinstance(value, str):
                value = str(value)

            string = string.replace(f"<<{key}>>", value.strip())

        return string

    @staticmethod
    def get_file_path_via_terminal() -> Optional[str]:
        root = tk.Tk()
        root.withdraw()
        file_path = filedialog.askopenfilename()
        root.destroy()
        if file_path == "":
            return None
        return file_path

class ObjectService:
    @staticmethod
    def validate_keys(data: Optional[Dict[str, Any]], keys: List[str]) -> bool:
        if data is None:
            data = {}

        assert isinstance(data, dict), f"data must be a dictionary. Value: {data}"
        assert isinstance(keys, list), f"keys must be a list. Value: {keys}"
        assert all(isinstance(key, str) for key in keys), f"all keys must be strings. Value: {keys}"

        return not keys or all(key in data for key in keys)

    @staticmethod
    def wrap(data: Any, key: Optional[str]) -> Union[Dict[str, Any], Any]:
        if not key:
            return data

        if isinstance(data, dict) and len(data) == 1 and key in data:
            return data

        return {key: data}

    @staticmethod
    def keys(data: Dict[Any, Any]) -> List[Any]:
        assert isinstance(data, dict), "Input object must be a dictionary"
        return list(data.keys())